//
// *****************************************************************************
// |          G4LECS -- GEANT4 Low-Energy Compton Scattering package.          |
// |                          R. M. Kippen / LANL                              |
// |                                                                           |
// | G4LECSRaylData.cc                                                         |
// |                                                                           |
// | Classes to read, store, and retrieve tabulated data for Rayleigh          |
// | scattering.                                                               |
// |                                                                           |
// | RMK, DEC-2003. Change g4std headers to standard gcc for G4.6.0            |
// | RMK, MAY-2002. First version (based on G3LECS & G4CompositeEMDataSet.cc)  |
// *****************************************************************************
//

#include "G4LECSRaylData.hh"
#include "Randomize.hh"
#include <fstream>
#include <strstream>


// *****************************************************************************
G4LECSRaylData::G4LECSRaylData(G4int minZ, G4int maxZ)
  :zMin(minZ), zMax(maxZ)
// *****************************************************************************
{
  nComponents = 0;
}

// *****************************************************************************
G4LECSRaylData::G4LECSRaylData(const G4String& dataFile,
					   G4int minZ, G4int maxZ)
  : zMin(minZ), zMax(maxZ)
// *****************************************************************************
{
  nComponents = 0;
  LoadData(dataFile);

}

// *****************************************************************************
G4LECSRaylData::~G4LECSRaylData()
// *****************************************************************************
{ 
  for (size_t i=0; i<nComponents; i++) 
    {
      delete components[i];
    }
}

// *****************************************************************************
G4double G4LECSRaylData::MaxTabulatedQval(G4int Z) const
// *****************************************************************************
{
  // Return the maximum tabulated momentum transfer
  // Q = 2 * E * sin(th/2) in MeV for this element Z.
  //
  G4int zIndex = Z-1;
  return components[zIndex]->rayQ[components[zIndex]->nPoints-1];

}

// *****************************************************************************
G4double G4LECSRaylData::RandomQval(G4int Z, G4double Qmax) const
// *****************************************************************************
{
  // Randomly select momentum transfer Q = 2 * E * sin(th/2) in MeV
  // for element Z according to the normalized cumulative 
  // form factor distribution for this element.  The Q value is 
  // interpolated from tabulated form factor CDF data.  The output
  // value will range from 0-->Qmax
  //
  G4int zIndex = Z-1;

  // CDF corresponding to Qmax
  G4double CDFmax =  LogLogIntrp(Qmax,
				 components[zIndex]->rayQ,
				 components[zIndex]->rayCDF);

  // Random sample of Q from CDF(0-->CDFmax)
  return LogLogIntrp(CDFmax*G4UniformRand(),
		     components[zIndex]->rayCDF, 
		     components[zIndex]->rayQ);
}

// *****************************************************************************
void G4LECSRaylData::PrintData() const
// *****************************************************************************
{
  G4cout << "The full data set has " << nComponents << " components" << G4endl;
  
  for (size_t i=0; i<nComponents; i++)
    {
      G4cout << "--- Component Z = " << i+1 << " ---" << G4endl;
      for (size_t j=0; j<components[i]->nPoints; j++) {
	G4cout << components[i]->rayQ[j] << "   "
	       << components[i]->rayCDF[j] << G4endl;
      }
    }

}

// *****************************************************************************
void G4LECSRaylData::LoadData(const G4String& fileName)
// *****************************************************************************
{ 

  // Loop over all files, one for each element Z in specified range
  //
  for (G4int Z=zMin; Z<=zMax; Z++) {

    // Build the complete string identifying the file with the data set
    //      
    char nameChar[100] = {""};
    std::ostrstream ost(nameChar, 100, std::ios::out);      
    ost << fileName << Z << ".dat";      
    G4String name(nameChar);      
    char* path = getenv("G4LEDATA");
    if (!path) { 
      G4String excep = "G4LECSRaylData - G4LEDATA environment variable not set";
      G4Exception(excep);
    }     

    // Open file stream in proper directory
    //
    G4String pathString(path);
    G4String dirFile = pathString + "/" + name;
    std::ifstream file(dirFile);
    std::filebuf* lsdp = file.rdbuf();      
    if (! (lsdp->is_open()) ) {
      G4String excep = "G4LECSRaylData - data file: " + dirFile + " not found";
      G4Exception(excep);
    }
    
    // Read and temporarily save data
    // Convert X=sin(t/2)/Wavel (1/cm) to Q = 2 * E *sin(t/2) (MeV)
    //
    G4double a = 0.0;
    G4int k = 1;
    G4DataVector qTMP, yTMP;
    do {
      file >> a;
      G4int nColumns = 2;
      // The file is organized into two columns:
      // 1st column is the X=sin(th/2)/Wavelength (1/cm)
      // 2nd column is the corresponding form factor (0-->Z^2)
      // The file terminates with the pattern: -1   -1
      //                                       -2   -2
      if (a == -1 || a == -2) {
      } else {
	if (k%nColumns != 0) {	
	  G4double q = a * 2.0*h_Planck*c_light/cm;
	  qTMP.push_back(q);
	  k++;
	} else if (k%nColumns == 0) {
	  G4double y = a;
	  yTMP.push_back(y);
	  k = 1;
	}
      }
    } while (a != -2); // end of file
    file.close();      

    // Initialize new elemental data object
    //
    G4LECSElemData* eleDat = new G4LECSElemData();      
    size_t i;

    // Redefine form factor for CDF calculation (FF^2 * Q)
    //
    for (i=0; i<yTMP.size(); i++) 
      yTMP[i] = yTMP[i]*yTMP[i] * qTMP[i];

    // Define a fine log-spaced momentum transfer (Q) grid for the
    // cumulative CDF.  Q = 2*E*sin(t/2) is in units of MeV
    //
    eleDat->nPoints = 201;
    eleDat->rayQ.resize(eleDat->nPoints);
    G4double Qmin = qTMP[1];              
    G4double Qmax = qTMP[qTMP.size()-1];
    G4double lStrt = log10(Qmin);
    G4double lStep = log10(Qmax/Qmin)/(eleDat->nPoints - 2);
    eleDat->rayQ[0] = 0.0;                   // Special first point
    for (i=1; i<eleDat->nPoints; i++) 
      eleDat->rayQ[i] = pow(10, lStrt+(i-1)*lStep);
    
    // Compute cumulative CDF (normalized [0->1] over the new Q grid 
    // N.B. CDF == Int{FF^2 dq^2} = Int{FF^2 * 2q dq}
    //
    eleDat->rayCDF.resize(eleDat->nPoints);
    eleDat->rayCDF[0] = 0.0;      // CDF(0.0) :== 0.0 (integral from 0-->0 = 0)
    for (i=1; i<eleDat->nPoints; i++)  {
      eleDat->rayCDF[i] = eleDat->rayCDF[i-1] + 
	TableInteg(eleDat->rayQ[i-1], eleDat->rayQ[i], qTMP, yTMP);
    }
    for (i=0; i<eleDat->nPoints; i++)      // Normalize
      eleDat->rayCDF[i] = eleDat->rayCDF[i]/eleDat->rayCDF[eleDat->nPoints - 1];

    components.push_back(eleDat);

    nComponents++;

  }  // End loop over elements

}

// *****************************************************************************
G4double G4LECSRaylData::LogLogIntrp(G4double xval,
					const G4DataVector& xdat, 
					const G4DataVector& ydat) const
// *****************************************************************************
{
  //
  // Interpolate value of tabulated data set at point X = xval.
  // Uses linear interpolation in Log-Log space for all but the first data
  // point, which, if X[0] = zero, will be linearly interpolated.
  //

  G4double value = 0.;
  G4int nDat = xdat.size();

  // Locate nearest values in xdat array using binary search
  //
  G4double xtmp = xval;
  if (xtmp < xdat[0]) xtmp = xdat[0];
  if (xtmp > xdat[nDat-1]) xtmp = xdat[nDat-1];
  size_t lowerBound = 0;
  size_t upperBound = nDat - 2; 
  while (lowerBound <= upperBound) {
      size_t midBin = (lowerBound + upperBound)/2;
      if ( xtmp < xdat[midBin] ) upperBound = midBin-1;
      else lowerBound = midBin+1;
  }  
  size_t bin = upperBound;

  // Interpolate value of Y between two nearest values
  //
  if (bin == 0 && (xtmp <= 0. || xdat[0] <= 0.) ) {
    G4double x1 = xdat[bin];
    G4double x2 = xdat[bin+1];              // If first bin is X = Zero then
    G4double y1 = ydat[bin];                // use linear interpolation
    G4double y2 = ydat[bin+1];
    value = y1 + (y2-y1) * (xtmp-x1) / (x2-x1);
  } else {
    G4double x1 = log10(xdat[bin]);
    G4double x2 = log10(xdat[bin+1]);       // Otherwise use Log-Log
    G4double y1 = log10(ydat[bin]);         // interpolation
    G4double y2 = log10(ydat[bin+1]);
    value = y1 + (y2-y1) * (log10(xtmp)-x1) / (x2-x1);
    value = pow(10,value);
  }

  return value;
}

// *****************************************************************************
G4double G4LECSRaylData::TableInteg(G4double xlim1, G4double xlim2,
					const G4DataVector& xdat, 
					const G4DataVector& ydat) const
// *****************************************************************************
{
  //
  // Integrate a tabulated function between limits xlim1-->xlim2 using 
  // the Newton-Cotes 5-point formula.  The tabulated function is evaluated
  // using linear interpolation in LOG-LOG space.
  //
  G4double dX = (xlim2-xlim1)/4.0;
  G4double y1 = LogLogIntrp(xlim1,      xdat, ydat);
  G4double y2 = LogLogIntrp(xlim1+  dX, xdat, ydat);
  G4double y3 = LogLogIntrp(xlim1+2*dX, xdat, ydat);
  G4double y4 = LogLogIntrp(xlim1+3*dX, xdat, ydat);
  G4double y5 = LogLogIntrp(xlim2,      xdat, ydat);

  G4double value = dX * (14.*(y1+y5) + 64.*(y2+y4) + 24.*y3) / 45. ;

  return value;
}
