//
// *****************************************************************************
// |          G4LECS -- GEANT4 Low-Energy Compton Scattering package.          |
// |                          R. M. Kippen / LANL                              |
// |                                                                           |
// | G4LECSRayleigh.cc                                                         |
// |                                                                           |
// | Classes to initialize and perform the Rayleigh (coherent) scattering      |
// | process.  Algorithms based on W.R. Nelson et al. The EGS4 Code System.    |
// | SLAC-Report-265, December 1985.                                           |
// |                                                                           |
// | *** Present code for unpolarized radiation only ***                       |
// |                                                                           |
// | RMK, MAY-2003. Make compatible with G4.5.1                                |
// |                   - Remove CutsPerMaterialWarning                         |
// |                   - Added cuts per region (G4MaterialCutsCouple)          |
// | RMK, MAY-2002. First version (based on G3LECS & G4LowEnergyRayleigh.cc)   |
// *****************************************************************************
//
#include "G4LECSRayleigh.hh"
#include "G4LECSRaylData.hh"
#include "Randomize.hh"
#include "G4ParticleDefinition.hh"
#include "G4Track.hh"
#include "G4Step.hh"
#include "G4ForceCondition.hh"
#include "G4Gamma.hh"
#include "G4Electron.hh"
#include "G4DynamicParticle.hh"
#include "G4VParticleChange.hh"
#include "G4ThreeVector.hh"
#include "G4VCrossSectionHandler.hh"
#include "G4CrossSectionHandler.hh"
#include "G4VEMDataSet.hh"
#include "G4VDataSetAlgorithm.hh"
#include "G4LogLogInterpolation.hh"
#include "G4MaterialCutsCouple.hh"
#include "G4UnitsTable.hh"

// *****************************************************************************
G4LECSRayleigh::G4LECSRayleigh(const G4String& processName)
  : G4VDiscreteProcess(processName),
    lowEnergyLimit(250*eV),             
    highEnergyLimit(100*GeV),
    intrinsicLowEnergyLimit(10*eV),
    intrinsicHighEnergyLimit(100*GeV)
// *****************************************************************************
{
  if (lowEnergyLimit < intrinsicLowEnergyLimit || 
      highEnergyLimit > intrinsicHighEnergyLimit)
    {
      G4Exception("G4LECSRayleigh::G4LECSRayleigh - energy limit outside intrinsic process validity range");
    }
  
  crossSectionHandler = new G4CrossSectionHandler();

  G4String formFactorFile = "rayl/re-ff-";
  formFactorData = new G4LECSRaylData(formFactorFile); 

  meanFreePathTable = 0;
}

// ***************************************************************************** 
G4LECSRayleigh::~G4LECSRayleigh()
// *****************************************************************************
{
  delete meanFreePathTable;
  delete crossSectionHandler;
  delete formFactorData;
}

// *****************************************************************************
void G4LECSRayleigh::BuildPhysicsTable(const G4ParticleDefinition&)
// *****************************************************************************
{
  
  crossSectionHandler->Clear();
  G4String crossSectionFile = "rayl/re-cs-";
  crossSectionHandler->LoadData(crossSectionFile);

  delete meanFreePathTable;
  meanFreePathTable = crossSectionHandler->BuildMeanFreePathForMaterials();

  PrintInfoDefinition();  
}

// *****************************************************************************
G4VParticleChange* G4LECSRayleigh::PostStepDoIt(const G4Track& aTrack, 
						const G4Step& aStep)
// *****************************************************************************
{
  // Simulates Rayleigh (i.e., coherent) scattering including atomic electron 
  // binding effects using tabulated & then interpolated form factor data for 
  // each element.
  //
  // The scattered gamma energy is sampled according to stored cumulative
  // Form Factors multiplied by the Rayleigh distribution.  Technique is
  // from EGS4 W.R. Nelson et al. The EGS4 Code System. SLAC-Report-265 , 
  // December 1985.
  //

  //-----------------------------------------------------------
  // Initialize dynamic particle quantities  
  //-----------------------------------------------------------
  aParticleChange.Initialize(aTrack);

  const G4DynamicParticle* incidentPhoton = aTrack.GetDynamicParticle();
  G4double photonEnergy0 = incidentPhoton->GetKineticEnergy();
  
  if (photonEnergy0 <= lowEnergyLimit)
    {
      aParticleChange.SetStatusChange(fStopAndKill);
      aParticleChange.SetEnergyChange(0.);
      aParticleChange.SetLocalEnergyDeposit(photonEnergy0);
      return G4VDiscreteProcess::PostStepDoIt(aTrack,aStep);
    }

  G4ParticleMomentum photonDirection0 = incidentPhoton->GetMomentumDirection();

  //-----------------------------------------------------------
  // Randomly select one element (Z) in the current material
  //-----------------------------------------------------------
  const G4MaterialCutsCouple* couple = aTrack.GetMaterialCutsCouple();
  G4int Z = crossSectionHandler->SelectRandomAtom(couple,photonEnergy0);

  //-----------------------------------------------------------
  //  Sample the zenith scatter angle using form factor FF
  //  dsigma/dOmega = ro^2/2 * (1 + cos^2[Th]) * FF(q,Z)^2
  //-----------------------------------------------------------
  G4double gReject;
  G4double cosTheta;

  do {
    // Max possible momentum transfer Q [i.e., 180 deg]
    G4double Qmax = 2. * photonEnergy0;

    // Randomly sample momentum transfer Q in MeV
    G4double randomQ = formFactorData->RandomQval(Z,Qmax);

    // Rejection function
    G4double QoverE = randomQ/photonEnergy0;
    cosTheta = 1. - 0.5*(QoverE*QoverE);
    gReject = 0.5*(1. + cosTheta*cosTheta);   

  } while( gReject < G4UniformRand() );

  G4double sinTheta = sqrt( (1.+cosTheta) * (1.-cosTheta) );
  
  //-----------------------------------------------------------
  // Scatter direction with uniform azimuth angle (unpolarized)
  //-----------------------------------------------------------
  G4double phi = twopi * G4UniformRand() ;
  G4double dirX = sinTheta*cos(phi);
  G4double dirY = sinTheta*sin(phi);
  G4double dirZ = cosTheta;
  
  //-----------------------------------------------------------
  // Update G4VParticleChange for the scattered photon 
  //-----------------------------------------------------------  
  G4ThreeVector photonDirection1(dirX, dirY, dirZ);

  photonDirection1.rotateUz(photonDirection0);

  aParticleChange.SetEnergyChange(photonEnergy0);
  aParticleChange.SetMomentumChange(photonDirection1);  
  aParticleChange.SetNumberOfSecondaries(0);

  return G4VDiscreteProcess::PostStepDoIt(aTrack, aStep);
}

// *****************************************************************************
G4bool G4LECSRayleigh::IsApplicable(const G4ParticleDefinition& particle)
// *****************************************************************************
{
  return ( &particle == G4Gamma::Gamma() ); 
}

// *****************************************************************************
G4double G4LECSRayleigh::GetMeanFreePath(const G4Track& track, 
					 G4double, 
					 G4ForceCondition*)
// *****************************************************************************
{
  const G4DynamicParticle* photon = track.GetDynamicParticle();
  G4double energy = photon->GetKineticEnergy();
  const G4MaterialCutsCouple* couple = track.GetMaterialCutsCouple();
  size_t materialIndex = couple->GetIndex();

  G4double meanFreePath;
  if (energy > highEnergyLimit) meanFreePath = meanFreePathTable->FindValue(highEnergyLimit,materialIndex);
  else if (energy < lowEnergyLimit) meanFreePath = DBL_MAX;
  else meanFreePath = meanFreePathTable->FindValue(energy,materialIndex);
  return meanFreePath;
}


// *****************************************************************************
void G4LECSRayleigh::PrintInfoDefinition()
// *****************************************************************************
{
                     
  G4cout << G4endl << GetProcessName() << ":  "
	 << "Rayleigh scattering with tabulated form factors.\n"
         << "        PhysicsTables from "
	           << G4BestUnit(lowEnergyLimit,"Energy")
         << " to " << G4BestUnit(highEnergyLimit,"Energy") << G4endl;

}         
